/*
 nasm -f elf busasm.asm   for busasm.o 
 gcc busspeed.c cpuidc.c  busasm.o -lrt -lc -lm -O3 -o busspeed32
 ./busspeed32

 nasm -f elf64 -g -F stabs busasm64.asm   for busasm64.o 
 gcc busspeed.c cpuidc64.c  busasm64.o -m64 -lrt -lc -lm -O3 -o busspeed64
 ./busspeed64

 *  BusSpeed
 *
 */

 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include "cpuidh.h"
 #include <math.h>
 #include <inttypes.h>
 #include <stdint.h>
 #include <malloc.h>
 #include <mm_malloc.h>

 


//  ************** PROTOPTYPES *****************

void adjust();


void runAll();

void inc32words();
void inc16words();
void inc8words();
void inc4words();
void inc2words();
void inc1word();
void incSSE2();

  #define Bits64
// #define Bits32
#ifdef Bits64
#else
#endif
 
// ************* GLOBAL VARIABLES **********

#ifdef Bits64
 long long    * array;
 long long    andsum1 = 0x5555555555555555;
 long long    andsumx = 0x5555555555555555;
 unsigned int wordBytes = 8;
 char   version[30] = "64 bit Version 2.0";
 char mips[80] = " Speed in MB/Second - For 64 bit MIPS divide by 8"; 
#else
 int          * array;
 int          andsum1 = 0x55555555;
 int          andsumx = 0x55555555;
 unsigned int wordBytes = 4;
 char   version[30] = "32 bit Version 2.0";
 char mips[80] = " Speed in MB/Second - For 32 bit MIPS divide by 4"; 
#endif

int           * andarray;

int            sse2loopinc;
long long    * arraym;     // was __m128i
long long    * andSSE2;    // was __m128i

int             memoryWords[20];
unsigned int    memoryKBytes[20];
double          memoryMB[20];
unsigned int    ramKB;
unsigned int    ramMB;

char    writeMsg[200];

int  runStage = -2;
int  runStage2 = 0;
int  mbpersec[20][20];


char fileName[256] = "BusSpd.txt";


volatile int testToRun;
volatile int wordsToTest;
volatile int passes1;
volatile double secs1;
volatile double secs2;
int runPasses = 10;

int passes1Loops[20][20];
int msecs1[30][20];
int passesa[20][20];
int passesb[20][20];

int     loopx;
double  passCount;
 
int ipass;


double runSecs = 0.05;


int  main()
{
   int i, g, p;
   int j, tt0, tt7;
   int endTest;
   int errors = 0;
   
   double addressInc[10];
   FILE *outfile;
   printf("\n");
   
   local_time();
   outfile = fopen(fileName,"a+");
   if (outfile == NULL)
   {
        printf ("Cannot open results file \n\n");
        printf(" Press Enter\n");
        g  = getchar();
        exit (0);
   }
   getDetails();
   printf("\n");
   for (i=1; i<10; i++)
   {
       printf("%s\n", configdata[i]);
   }
   fprintf (outfile, " #####################################################################\n");
   for (i=1; i<10; i++)
   {
       fprintf(outfile, "%s \n", configdata[i]);
   }
   printf("\n");
   printf(" ##########################################\n"); 
   printf ("         Bus Speed Test %s %s \n", version, timeday);
   printf ("   Kbytes Inc32wds Inc16wds  Inc8wds  Inc4wds  Inc2wds  ReadAll 128bSSE2\n\n");


   fprintf (outfile, " #####################################################################\n");
   fprintf (outfile, "       Bus Speed Test %s %s \n", version, timeday);

   fprintf (outfile, "                    Copyright Roy Longbottom 2010\n\n");
   fprintf (outfile, "   Kbytes Inc32wds Inc16wds  Inc8wds  Inc4wds  Inc2wds  ReadAll 128bSSE2\n\n");
        //             12345678 12345678 12345678 12345678 12345678 12345678 12345678 12345678 

   fflush(outfile);

   addressInc[0] = 32; // inc32words
   addressInc[1] = 16; // inc16words
   addressInc[2] = 8;  // inc8words
   addressInc[3] = 4;  // inc4words
   addressInc[4] = 2;  // inc2words
   addressInc[5] = 1;  // inc1word
   addressInc[6] = 1;  // inc all words incSSE2
   
   memoryKBytes[0] = 6;
   memoryKBytes[1] = 24;
   memoryKBytes[2] = 96;
   memoryKBytes[3] = 384;
   memoryKBytes[4] = 768;
   memoryKBytes[5] = 1536;
   memoryKBytes[6] = 16380;  //  16384;
   memoryKBytes[7] = 131070; // 131072;
   memoryKBytes[8] = 393210; // 393216

   for (i=0; i<20; i++)
   {
       for (j=0; j<20; j++)
       {
           mbpersec[i][j] = 0;
           passes1Loops[j][i] = 0;
       }
   }
   ipass = 9;
   ramKB = (unsigned int)(ramGB * 1000000);
   ramMB = ramKB / 1024;
   if (ramMB < 1024) ipass = 8; 

   if (ramMB < 384 ) ipass = 7;
   if (ramMB < 32  ) ipass = 6;
   for (i=0; i<ipass; i++)
   {
       memoryWords[i]  = (int)((double)memoryKBytes[i] * 1024.0/ (double)wordBytes) ;
       memoryMB[i] = ((double)memoryKBytes[i] * 1024.0)/ 1000000;        
   }

#ifdef Bits64
   array = (long long *)_mm_malloc(memoryKBytes[ipass-1]*1024, 16);
   if (array == NULL)
   {
       printf(" Cannot allocate memory ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       g  = getchar();
       exit(1);
   }
#else
   array = (int *)_mm_malloc(memoryKBytes[ipass-1]*1024, 16);
   if (array == NULL)
   {
       printf(" Cannot allocate memory ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       g  = getchar();
       exit(1);
   }
 #endif
    arraym  = (long long  *)array; // was __m128i 

   

   andarray = (int *)_mm_malloc(16, 16);
   if (andarray == NULL)
   {
       printf(" Cannot allocate memory ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       _mm_free(array);
       g  = getchar();
       exit(1);
   }
   andSSE2 = (long long  *)andarray;  // was __m128i 

   for (i=0; i<memoryWords[ipass-1]; i++) // arraym
   {
       array[i] = andsum1;
   }


   for (j=0; j<4; j++) // andSSE2
   {
       andarray[j] = 0x55555555;
   }

   
   runStage = 0;
   runStage2 = 0;

   endTest = 7;
   if (hasSSE2) endTest = 8;
   
   for (i=0; i<ipass; i++)
   {
       wordsToTest = memoryWords[i];
       runStage = runStage + 1;
       printf(" %8d", memoryKBytes[i]);

       for (testToRun=1; testToRun<endTest; testToRun++)
       {
           tt0 = testToRun - 1;
           passes1 = 1;
           runAll();
           for (loopx=0; loopx<10; loopx++)
           {
               start_time();
               runAll();
               end_time();
               if (loopx == 0)
               {
                   msecs1[tt0 + 14][i] = (int)(secs * 1000000);
               }
               adjust();
           }
           passes1Loops[i][tt0] = passes1;
           start_time();
           for (p=0; p<runPasses; p++)
           {
               runAll();
           }
           end_time();
           
           mbpersec[tt0][i] = (int)((double)passes1 * (double)runPasses * memoryMB[i] / secs / addressInc[tt0]);
           msecs1[tt0][i] = (int)(secs * 1000);
           passesa[tt0][i] = passes1;
           printf(" %8d", mbpersec[tt0][i]);           
           fflush(stdout);
       }
      

       if (andsum1 != andsumx)
       {
#ifdef Bits64
           sprintf(writeMsg, "Wrong result Expected 0x%16.16lx Was 0x%16.16lx at %d KB", (long unsigned int)andsumx, (long unsigned int)andsum1, memoryKBytes[i]);
#else
           sprintf(writeMsg, "Wrong result Expected 0x%8.8x Was 0x%8.8x at %d KB", andsumx, andsum1, memoryKBytes[i]);
#endif
           fprintf(outfile, " %s\n", writeMsg);
           andsum1 = andsumx;
           errors = errors + 1;
       }

       if (hasSSE2)
       {
           for (j=0; j<4; j++) // andSSE2
           {
               if (andarray[j]  != 0x55555555)
               {
                   sprintf(writeMsg, "Wrong result Expected 0x55555555 Was 0x%8.8x at %d KB SSE2", andarray[j], memoryKBytes[i]);
                   fprintf(outfile, " %s\n", writeMsg);
                   andarray[j] = 0x55555555;
                   errors = errors + 1;
               }
           }

       }
       printf("\n");   
   }

   if (errors) printf("\n %d Sumcheck Errors - See Log\n", errors);

   printf("\n");   
   for (i=0; i<runStage; i++)
   {
       fprintf(outfile, " %8d %8d %8d %8d %8d %8d %8d %8d\n",        
                           memoryKBytes[i],
                           mbpersec[0][i], mbpersec[1][i],
                           mbpersec[2][i], mbpersec[3][i],
                           mbpersec[4][i], mbpersec[5][i],
                           mbpersec[6][i]);
   }
   fprintf (outfile, "\n"); 
   local_time();
   fprintf (outfile, "           %s\n\n", mips);
   fprintf (outfile, "                   End at %s\n", timeday);

   fprintf(outfile, "\n");
   fflush (outfile);
   fclose(outfile);

   testToRun = 0;
   wordsToTest = 0;

   _mm_free(andarray);  
   _mm_free(array);
   
   return 1;

} // runAllTests


void adjust()
{
   double runtime = runSecs;
   
   if (secs > runtime)
   {
       loopx = 11;
   }
   else if (secs < runtime / 100)
   {
       passes1 = passes1 * 100;
   }
   else if (secs > runtime / 2)
   {
       passes1 = passes1 * (int)(runtime / secs);
       loopx = 11;
   }
   else
   {
       passes1 = passes1 * (int)(runtime / secs);
   }
}

void runAll()
{
    if (testToRun == 1) inc32words();
    if (testToRun == 2) inc16words();
    if (testToRun == 3) inc8words();
    if (testToRun == 4) inc4words();
    if (testToRun == 5) inc2words();
    if (testToRun == 6) inc1word();
    if (testToRun == 7) incSSE2();   
    return;
}


void inc32words()
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+768)
       {
           andsum1 = andsum1 & array[i     ] & array[i+32  ] & array[i+64  ] & array[i+96  ]
                             & array[i+128 ] & array[i+160 ] & array[i+192 ] & array[i+224 ]
                             & array[i+256 ] & array[i+288 ] & array[i+320 ] & array[i+352 ]
                             & array[i+384 ] & array[i+416 ] & array[i+448 ] & array[i+480 ]
                             & array[i+512 ] & array[i+544 ] & array[i+576 ] & array[i+608 ]
                             & array[i+640 ] & array[i+672 ] & array[i+704 ] & array[i+736 ];
       }
   }
}

void inc16words()
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+768)
       {
           andsum1 = andsum1 & array[i    ] & array[i+16 ] & array[i+32 ] & array[i+48 ]
                             & array[i+64 ] & array[i+80 ] & array[i+96 ] & array[i+112]
                             & array[i+128] & array[i+144] & array[i+160] & array[i+176]
                             & array[i+192] & array[i+208] & array[i+224] & array[i+240]
                             & array[i+256] & array[i+272] & array[i+288] & array[i+304]
                             & array[i+320] & array[i+336] & array[i+352] & array[i+368]
                             & array[i+384] & array[i+400] & array[i+416] & array[i+432]
                             & array[i+448] & array[i+464] & array[i+480] & array[i+496]
                             & array[i+512] & array[i+528] & array[i+544] & array[i+560]
                             & array[i+576] & array[i+592] & array[i+608] & array[i+624]
                             & array[i+640] & array[i+656] & array[i+672] & array[i+688]
                             & array[i+704] & array[i+720] & array[i+736] & array[i+752];
       }
   }
}

void inc8words()
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+384)
       {
           andsum1 = andsum1 & array[i    ] & array[i+8  ] & array[i+16 ] & array[i+24 ]
                             & array[i+32 ] & array[i+40 ] & array[i+48 ] & array[i+56 ]
                             & array[i+64 ] & array[i+72 ] & array[i+80 ] & array[i+88 ]
                             & array[i+96 ] & array[i+104] & array[i+112] & array[i+120]
                             & array[i+128] & array[i+136] & array[i+144] & array[i+152]
                             & array[i+160] & array[i+168] & array[i+176] & array[i+184]
                             & array[i+192] & array[i+200] & array[i+208] & array[i+216]
                             & array[i+224] & array[i+232] & array[i+240] & array[i+248]
                             & array[i+256] & array[i+264] & array[i+272] & array[i+280]
                             & array[i+288] & array[i+296] & array[i+304] & array[i+312]
                             & array[i+320] & array[i+328] & array[i+336] & array[i+344]
                             & array[i+352] & array[i+360] & array[i+368] & array[i+376];
       }
   }
}

void inc4words()
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+256)
       {
           andsum1 = andsum1 & array[i    ] & array[i+4  ] & array[i+8  ] & array[i+12 ]
                             & array[i+16 ] & array[i+20 ] & array[i+24 ] & array[i+28 ]
                             & array[i+32 ] & array[i+36 ] & array[i+40 ] & array[i+44 ]
                             & array[i+48 ] & array[i+52 ] & array[i+56 ] & array[i+60 ]
                             & array[i+64 ] & array[i+68 ] & array[i+72 ] & array[i+76 ]
                             & array[i+80 ] & array[i+84 ] & array[i+88 ] & array[i+92 ]
                             & array[i+96 ] & array[i+100] & array[i+104] & array[i+108]
                             & array[i+112] & array[i+116] & array[i+120] & array[i+124]
                             & array[i+128] & array[i+132] & array[i+136] & array[i+140]
                             & array[i+144] & array[i+148] & array[i+152] & array[i+156]
                             & array[i+160] & array[i+164] & array[i+168] & array[i+172]
                             & array[i+176] & array[i+180] & array[i+184] & array[i+188]
                             & array[i+192] & array[i+196] & array[i+200] & array[i+204]
                             & array[i+208] & array[i+212] & array[i+216] & array[i+220]
                             & array[i+224] & array[i+228] & array[i+232] & array[i+236]
                             & array[i+240] & array[i+244] & array[i+248] & array[i+252];
       }
   }
}

void inc2words()
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+128)
       {
           andsum1 = andsum1 & array[i    ] & array[i+2  ] & array[i+4  ] & array[i+6  ]
                             & array[i+8  ] & array[i+10 ] & array[i+12 ] & array[i+14 ]
                             & array[i+16 ] & array[i+18 ] & array[i+20 ] & array[i+22 ]
                             & array[i+24 ] & array[i+26 ] & array[i+28 ] & array[i+30 ]
                             & array[i+32 ] & array[i+34 ] & array[i+36 ] & array[i+38 ]
                             & array[i+40 ] & array[i+42 ] & array[i+44 ] & array[i+46 ]
                             & array[i+48 ] & array[i+50 ] & array[i+52 ] & array[i+54 ]
                             & array[i+56 ] & array[i+58 ] & array[i+60 ] & array[i+62 ]
                             & array[i+64 ] & array[i+66 ] & array[i+68 ] & array[i+70 ]
                             & array[i+72 ] & array[i+74 ] & array[i+76 ] & array[i+78 ]
                             & array[i+80 ] & array[i+82 ] & array[i+84 ] & array[i+86 ]
                             & array[i+88 ] & array[i+90 ] & array[i+92 ] & array[i+94 ]
                             & array[i+96 ] & array[i+98 ] & array[i+100] & array[i+102]
                             & array[i+104] & array[i+106] & array[i+108] & array[i+110]
                             & array[i+112] & array[i+114] & array[i+116] & array[i+118]
                             & array[i+120] & array[i+122] & array[i+124] & array[i+126];
       }
    }
}

void inc1word()
{
   int i, j;

   for(j=0; j<passes1; j++)
   {
       for (i=0; i<wordsToTest; i=i+64)
       {
           andsum1 = andsum1 & array[i   ] & array[i+1 ] & array[i+2 ] & array[i+3 ]
                             & array[i+4 ] & array[i+5 ] & array[i+6 ] & array[i+7 ]
                             & array[i+8 ] & array[i+9 ] & array[i+10] & array[i+11]
                             & array[i+12] & array[i+13] & array[i+14] & array[i+15]
                             & array[i+16] & array[i+17] & array[i+18] & array[i+19]
                             & array[i+20] & array[i+21] & array[i+22] & array[i+23]
                             & array[i+24] & array[i+25] & array[i+26] & array[i+27]
                             & array[i+28] & array[i+29] & array[i+30] & array[i+31]
                             & array[i+32] & array[i+33] & array[i+34] & array[i+35]
                             & array[i+36] & array[i+37] & array[i+38] & array[i+39]
                             & array[i+40] & array[i+41] & array[i+42] & array[i+43]
                             & array[i+44] & array[i+45] & array[i+46] & array[i+47]
                             & array[i+48] & array[i+49] & array[i+50] & array[i+51]
                             & array[i+52] & array[i+53] & array[i+54] & array[i+55]
                             & array[i+56] & array[i+57] & array[i+58] & array[i+59]
                             & array[i+60] & array[i+61] & array[i+62] & array[i+63];
       }
   }
}

void incSSE2()
{
   int j;
   sse2loopinc = wordsToTest * wordBytes / 1024;
   for(j=0; j<passes1; j++)
   {
      _asmsse2();
   }
}



