/* 
 nasm -f elf64 -g -F stabs sse64.asm   ;for sse64.o
 gcc SSEfpu.c cpuidc64.c  sse64.o -lrt -lc -lm -O3 -o ssefpu64
 ./ssefpu64

 nasm -f elf sse32.asm   ;for sse32.o 
 gcc SSEfpu.c cpuidc.c  sse32.o -lrt -lc -lm -O3 -o ssefpu32
 ./ssefpu32

*/

 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include "cpuidh.h"
 #include <math.h>
 #include <malloc.h>
 #include <mm_malloc.h>

double * xd;
double * yd;
float  * xs;
float  * ys;
double * dsum;
float  * ssum;
int kd, ks;
int kd1, ks2;

FILE    *outfile;


//  ************** PROTOPTYPES *****************

void checkTime();

// ************* GLOBAL VARIABLES **********
 
 char   version[30] = "64-Bit Version 2.1";
// char   version[30] = "32-Bit Version 2.1";

#define BITS32   // See later needed to limit allocated memory
 
int     n1;
double  runTime = 0.1;

double results[30][9];
double maxmbps[8];

int runs = 10; // 2 MB;
int passes[25];
int allocMem[25];


int main()
{

    float  sums;
    float  ones = 1.0;
    float  zeros = 0;
    int i, g, m, j, nn;
    int mem;
    int inc;
    double BytesPerFLOP;
    double sumd;
    double  ramKB;
    long long useMem;
    double memMB;
    
    double oned = 1.0;
    double zerod = 0;

    outfile = fopen("SSEfpu.txt","a+");
    if (outfile == NULL)
    {
        printf ("Cannot open results file \n\n");
        printf(" Press Enter\n");
        g  = getchar();
        exit (0);
    }
    printf("\n");
    getDetails();

    for (i=1; i<10; i++)
    {
        printf("%s\n", configdata[i]);
    }
    printf("\n");

    fprintf (outfile, " ###############################################################\n\n");                     
    for (i=1; i<10; i++)
    {
        fprintf(outfile, "%s \n", configdata[i]);
    }
    fprintf (outfile, "\n");
 
    printf(" ##########################################\n"); 
    fprintf (outfile, " ###############################################################\n");                     

    ramKB = ramGB * 1000000;

    allocMem[0] = 2;  // KB two arrays
    for (i=1; i<23; i++)
    {
        allocMem[i] = allocMem[i-1] * 2;
    }
 
    passes[0] = 256;        //      4 KB
    passes[1] = 512;        //      8 KB
    passes[2] = 1024;       //     16 KB 
    passes[3] = 2000;       //     32 KB
    passes[4] = 4000;       //     64 KB
    passes[5] = 8000;       //    128 KB
    passes[6] = 16000;      //    256 KB
    passes[7] = 32000;      //    512 KB
    passes[8] = 64000;      //    1 MB
    passes[9] = 128000;     //    2 MB
    passes[10] = 256000;    //    4 MB
    passes[11] = 512000;    //    8 MB
    passes[12] = 1024000;   //   16 MB
    passes[13] = 2048000;   //   32 MB
    passes[14] = 4096000;   //   64 MB
    passes[15] = 8192000;   //  128 MB
    passes[16] = 16384000;  //  256 MB
    passes[17] = 32768000;  //  512 MB
    passes[18] = 65536000;  // 1024 MB
    passes[19] = 131072000; // 2048 MB
    passes[20] = 262144000; // 4096 MB
    passes[21] = 524288000; // 8192 MB
    passes[22] =1048576000; //16384 MB

    if (ramKB >    14000) runs = 11;
    if (ramKB >    30000) runs = 12;
    if (ramKB >    60000) runs = 13;
    if (ramKB >   120000) runs = 14;
    if (ramKB >   250000) runs = 15;
    if (ramKB >   500000) runs = 16;
    if (ramKB >  1000000) runs = 17;
    if (ramKB >  2000000) runs = 18;
    if (ramKB >  4000000) runs = 19;
    if (ramKB >  8000000) runs = 20;
    if (ramKB > 16000000) runs = 21;
    if (ramKB > 32000000) runs = 22;
    if (ramKB > 64000000) runs = 23;
    
    #ifdef  BITS32
     if (runs > 19) runs = 19;
    #endif
    for (i=0; i<30; i++)
    {
        for (j=0; j<9; j++)
        {
            results[i][j] = 0;
        }
    }
    useMem = allocMem[runs - 1]; // passes[runs - 1] * 8 + 8;    
    xd = (double *)_mm_malloc(useMem * 1024 + 64, 16);
    if (xd == NULL)
    {
       printf(" ERROR WILL EXIT\n");
       printf(" Press Enter\n");
       g  = getchar();
       exit(1);
    }

    yd = (double *)_mm_malloc(useMem * 1024 + 64, 16);
    if (yd == NULL)
    {
       printf(" ERROR WILL EXIT\n");
       printf(" Press Enter\n");
        _mm_free(xd);
       g  = getchar();
       exit(1);
    }

    dsum = (double *)_mm_malloc(128, 16);
    if (dsum == NULL)
    {
        _mm_free(xd);
        _mm_free(yd);
       g  = getchar();
       exit(1);
    }        

    local_time();

    printf("\n     SSE & SSE2 Memory Reading Speed Test %s\n", version);      
    printf("\n               Copyright (C) Roy Longbottom 2010\n\n");    
    printf("  Memory    --s=s+x[m]*y[m]---   --x[m]=x[m]+y[m]-- (s+x[m])?y[m]\n"); 

    fprintf(outfile, "\n     SSE & SSE2 Memory Reading Speed Test %s\n", version);      
    fprintf(outfile, "\n               Copyright (C) Roy Longbottom 2010\n");    
    fprintf(outfile, "\n     %6.3f seconds per test, Start %s\n", runTime, timeday);
    fprintf(outfile, "  Memory    --s=s+x[m]*y[m]---   --x[m]=x[m]+y[m]-- (s+x[m])?y[m]\n"); 
    fflush(outfile);

    xs   = (float  * )xd;
    ys   = (float  * )yd;
    ssum = (float  * )dsum;

    printf("  KBytes    SSE2    SSE   Sngl   SSE2    SSE   Sngl  +*SSE  ++SSE\n");   
    printf("   Used     MB/S   MB/S   MB/S   MB/S   MB/S   MB/S   MB/S   MB/S\n\n");

    fprintf(outfile, "  KBytes    SSE2    SSE   Sngl   SSE2    SSE   Sngl  +*SSE  ++SSE\n");   
    fflush(outfile);

    fprintf(outfile, "   Used     MB/S   MB/S   MB/S   MB/S   MB/S   MB/S   MB/S   MB/S\n\n");
    fflush(outfile);

    for (i=0; i<8; i++)
    {
        maxmbps[i] = 0;
    }

    for (j=0; j<runs; j++)
    {
        kd = passes[j];
        nn = 6400000 / kd;  // min 240
        if (nn < 1) nn = 1;
        
        ks = kd * 2;
        kd1 = kd;

        memMB = (double)kd * 16.0 / 1000000.0;
        mem = (int)((double)kd * 16.0 / 1000.0);
       
        inc = 4;

        // SSE2One *******************************  


        if (hasSSE2)
        {
            n1 = nn;
      
           do
            {
                sumd = 0;
                dsum[0] = 0;
                dsum[1] = 0;
                for (m=0; m<kd1+6; m++)
                {
                      xd[m] = oned;
                      yd[m] = oned;
                }          
                start_time();  
                for (i=0; i<n1; i++)
                {
                    _SSE2One();
                }
                end_time();
                checkTime();
            }
            while (secs < runTime);
   
            sumd = dsum[0] + dsum[1];
            results[j][0] = (double)n1 * memMB / secs;
            printf("%8d %7d", mem, (int)results[j][0]);
            fflush(stdout);
            fprintf(outfile, "%8d ", mem);  
            BytesPerFLOP = 8;            
            if (results[j][0] / BytesPerFLOP > maxmbps[0]) maxmbps[0] = results[j][0] / BytesPerFLOP;
        }
    
        //  SSEOne ******************************* 

        if (hasSSE)
        {
            n1 = nn;
            do
              {
                ssum[0] = 0;
                ssum[1] = 0;
                ssum[2] = 0;
                ssum[3] = 0;
                for (m=0; m<ks+12; m++)
                {
                      xs[m] = ones;           
                      ys[m] = ones;
                }          
                start_time();
                
                for (i=0; i<n1; i++)
                  {
                    _SSEOne();
                  }
                end_time();
                checkTime();            
              }
            while (secs < runTime);
    
            sums = ssum[0] + ssum[1] + ssum[2] + ssum[3];        
            results[j][1] = (double)n1 * memMB / secs;
            printf("%7d", (int)results[j][1]);
            fflush(stdout);
            BytesPerFLOP = 4;            
            if (results[j][1] / BytesPerFLOP > maxmbps[1]) maxmbps[1] = results[j][1] / BytesPerFLOP;
        }


    //  SPOne ******************************* 

        n1 = nn;
        inc = 16;
        do
          {
            ssum[0] = 0;
            for (m=0; m<ks; m++)
              {
                  xs[m] = ones;
                  ys[m] = ones;
              }          
            start_time();
            
            for (i=0; i<n1; i++)
              {
                for (m=0; m<ks-inc; m=m+inc)
                 {
                   sums = sums + xs[m] * ys[m]
                               + xs[m+1]  * ys[m+1]
                               + xs[m+2]  * ys[m+2]
                               + xs[m+3]  * ys[m+3]                   
                               + xs[m+4]  * ys[m+4]
                               + xs[m+5]  * ys[m+5]
                               + xs[m+6]  * ys[m+6]                   
                               + xs[m+7]  * ys[m+7]
                               + xs[m+8]  * ys[m+8]
                               + xs[m+9]  * ys[m+9]                   
                               + xs[m+10] * ys[m+10]
                               + xs[m+11] * ys[m+11]
                               + xs[m+12] * ys[m+12]                   
                               + xs[m+13] * ys[m+13]
                               + xs[m+14] * ys[m+14]
                               + xs[m+15] * ys[m+15];                   
                 }
              }
            end_time();
            checkTime();            
          }
        while (secs < runTime);

        if (sums == 0.0) fprintf(outfile, " SPOne Wrong Result\n");         
        results[j][2] = (double)n1 * memMB / secs;
        printf("%7d", (int)results[j][2]);
        fflush(stdout);
        BytesPerFLOP = 4;            
        if (results[j][2] / BytesPerFLOP > maxmbps[2]) maxmbps[2] = results[j][2] / BytesPerFLOP;

    
        //  SSE2Two ******************************* 
        
        if (hasSSE2)
        {
            n1 = nn;
            do
              {
                for (m=0; m<kd; m++)
                  {
                      xd[m] = zerod;
                      yd[m] = oned;
                  }          
                start_time();
                for (i=0; i<n1; i++)
                  {
                    _SSE2Two();
                  }
                end_time();
                checkTime();
              }
            while (secs < runTime);
        
            sumd = xd[1];
            results[j][3] = (double)n1 * memMB / secs;
            printf("%7d", (int)results[j][3]);
            fflush(stdout);
            BytesPerFLOP = 16;            
            if (results[j][3] / BytesPerFLOP > maxmbps[3]) maxmbps[3] = results[j][3] / BytesPerFLOP;
        }
    
        //  SSETwo ******************************* 

        if (hasSSE)
        {
            n1 = nn;
            do
              {
                for (m=0; m<ks; m++)
                  {
                      xs[m] = zeros;
                      ys[m] = ones;
                  }          
                start_time();
                for (i=0; i<n1; i++)
                  {
                    _SSETwo();
                  }
                end_time();
                checkTime();
              }
            while (secs < runTime);
                
            sums = xs[1];
            results[j][4] = (double)n1 * memMB / secs;
            printf("%7d", (int)results[j][4]);
            fflush(stdout);
            BytesPerFLOP = 8;            
            if (results[j][4] / BytesPerFLOP > maxmbps[4]) maxmbps[4] = results[j][4] / BytesPerFLOP;
        }

 
        //  SPTwo ******************************* 

        n1 = nn;
        inc = 16;
        do
          {
            for (m=0; m<ks; m++)
              {
                  xs[m] = zeros;
                  ys[m] = ones;
              }          
            start_time();
            for (i=0; i<n1; i++)
              {
                for (m=0; m<ks-inc; m=m+inc)
                 {
                      xs[m]    = xs[m]    + ys[m];
                      xs[m+1]  = xs[m+1]  + ys[m+1];
                      xs[m+2]  = xs[m+2]  + ys[m+2];
                      xs[m+3]  = xs[m+3]  + ys[m+3];
                      xs[m+4]  = xs[m+4]  + ys[m+4];
                      xs[m+5]  = xs[m+5]  + ys[m+5];
                      xs[m+6]  = xs[m+6]  + ys[m+6];
                      xs[m+7]  = xs[m+7]  + ys[m+7];
                      xs[m+8]  = xs[m+8]  + ys[m+8];
                      xs[m+9]  = xs[m+9]  + ys[m+9];
                      xs[m+10] = xs[m+10] + ys[m+10];
                      xs[m+11] = xs[m+11] + ys[m+11];
                      xs[m+12] = xs[m+12] + ys[m+12];
                      xs[m+13] = xs[m+13] + ys[m+13];
                      xs[m+14] = xs[m+14] + ys[m+14];
                      xs[m+15] = xs[m+15] + ys[m+15];
                 }
              }
            end_time();
            checkTime();
          }
        while (secs < runTime);
            
        sums = xs[1];
        if (sums == 0.0) fprintf(outfile, " SPTwo Wrong Result\n");         
        results[j][5] =  (double)n1 * memMB / secs;
        printf("%7d", (int)results[j][5]);
        fflush(stdout);
        BytesPerFLOP = 8;            
        if (results[j][5] / BytesPerFLOP > maxmbps[5]) maxmbps[5] = results[j][5] / BytesPerFLOP;

        //  SSEOnem ******************************* 

        if (hasSSE)
        {   
            ks2 = ks/2;
            n1 = nn;
            do
              {
                for (m=0; m<32; m++)
                {
                   ssum[m] = 1.0;
                }
                for (m=0; m<ks; m++)
                {
                      xs[m] = ones;
                      ys[m] = ones;
                }          
                start_time();
                
                for (i=0; i<n1; i++)
                  {
                    _SSEOnem();
                  }
                end_time();
                checkTime();            
              }
            while (secs < runTime);
    
            sums = ssum[0] + ssum[1] + ssum[2] + ssum[3];        
            results[j][6] = (double)n1 * memMB / secs;
            printf("%7d", (int)results[j][6]);
            fflush(stdout);
            BytesPerFLOP = 4;            
            if (results[j][6] / BytesPerFLOP > maxmbps[6]) maxmbps[6] = results[j][6] / BytesPerFLOP;
        }

        //  SSEOnen ******************************* 

        if (hasSSE)
        {   
            ks2 = ks/2;
            n1 = nn;
            do
              {
                for (m=0; m<32; m++)
                {
                   ssum[m] = 1.0;
                }
                for (m=0; m<ks; m++)
                {
                      xs[m] = ones;
                      ys[m] = ones;
                }          
                start_time();
                
                for (i=0; i<n1; i++)
                  {
                    _SSEOnen();
                  }
                end_time();
                checkTime();            
              }
            while (secs < runTime);
    
            sums = ssum[0] + ssum[1] + ssum[2] + ssum[3];        
            results[j][7] = (double)n1 * memMB / secs;
            printf("%7d\n", (int)results[j][7]);
            fflush(stdout);
            BytesPerFLOP = 4;            
            if (results[j][7] / BytesPerFLOP > maxmbps[7]) maxmbps[7] = results[j][7] / BytesPerFLOP;
        }

        fprintf(outfile, "%7.0f%7.0f%7.0f",   results[j][0], results[j][1], results[j][2]);
        fprintf(outfile, "%7.0f%7.0f%7.0f",   results[j][3], results[j][4], results[j][5]);
        fprintf(outfile, "%7.0f%7.0f\n",      results[j][6], results[j][7]);

        fflush(outfile);
    }
    
    fprintf(outfile, "\n");
    fprintf(outfile, "            SSE2    SSE   Norm   SSE2    SSE   Norm    SSE    SSE\n");   
    fprintf(outfile, " Maximum      DP     SP     SP     DP     SP     SP     SP     SP\n");
    fprintf(outfile, "  MFLOPS ");
    for (i=0; i<8; i++)
    {
        fprintf(outfile, "%7.0f", maxmbps[i]);
    }
    fprintf(outfile, "\n");
    



    local_time();

    fprintf(outfile, "\n              End of test %s", timeday);        
    fprintf(outfile,"\n");
    
    fflush(outfile);
    fclose(outfile);
    _mm_free(xd);
    _mm_free(yd);
    _mm_free(dsum);
    printf("\n Press Enter\n");
    g  = getchar();

    return 1;
    
} // main


void checkTime()
{
    if (secs < runTime)
    {
          if (secs < runTime / 8.0)
          {
                n1 = n1 * 10;
          }
          else
          {
                n1 = (int)(runTime * 1.25 / secs * (double)n1+1);
           }
      }
}


         
  
